#!/bin/sh
# the next line restarts using tclsh \
exec tclsh "$0" "$@"

# this example shows a minimum application with menubar,
# statusbar, toolbar, and main window

load ../src/gnocl.so

# create submenu for menu "File" with standard items "New" and 
# "Quit" (text, icon and accelerator). 'N', 'Q' respectively, 
# are underlined and are used as mnemonics.
set menu [gnocl::menu]
$menu add [gnocl::menuItem label -text "%#New" -tooltip "Make new" \
      -command {puts "That's new"}]
$menu add [gnocl::menuItem separator]
$menu add [gnocl::menuItem label -text "%#Quit" -command exit \
      -tooltip "Quit program"]

# create menu "File", 'F' is underlined and used as mnemonic
set file [gnocl::menuItem label -text "%__File" -submenu $menu]

# create menu "Help" with item "About" 
set menu [gnocl::menu]
$menu add [gnocl::menuItem label -text "%__About" \
      -tooltip "Show about dialog" \
      -command {puts "Mini example (c) 2001 P.G. Baum"}]
set help [gnocl::menuItem label -text "%__Help" -submenu $menu]

# create toolbar with standard items "Quit" and "New"
set toolbar [gnocl::toolbar -style both]
$toolbar add item -text "%#Quit" -tooltip "Tooltip Quit" \
            -command exit
$toolbar add space
$toolbar add item -text "%#New" -tooltip "Tooltip new" \
   -command {puts "That's new"}

# create gtk application with menu, toolbar, statusbar 
# and a label as main widget
set box [gnocl::box -orientation vertical]
set win [gnocl::window -child $box -title "Test Application"]
$box add [gnocl::menubar -children [list $file $help]]  -expand 0 
$box add $toolbar -expand 0 
$box add [gnocl::label -text \
      {%<<span foreground="blue" size="large">Hello</span>\
      <span foreground="red" size="large">World</span>}]
$box add [gnocl::statusbar] -expand 0

if 0 {
# this is the gnome version
gnocl::app testApp -title "Test Application" -width 250 \
      -menubar [gnocl::menubar -children [list $file $help]] \
      -statusbar [gnocl::appBar] -toolbar $toolbar \
      -contents [gnocl::label -text "here comes the main widget"]
}


# if sourced from wish withdraw tk toplevel window, else
# enter gtk main loop
if { [info exists tk_patchLevel] } {
   wm withdraw .
} else {
   gnocl::mainLoop
}
