#!/bin/sh
# the next line restarts using tclsh \
exec tclsh "$0" "$@"

# this is an example for the text widget
# TODO: font, foreground and background

load ../src/gnocl.so

# create submenu for menu "File" 
# with standard items "New" and "Quit" (text and icon)
set menu [gnocl::menu]
$menu add [gnocl::menuItem label -text "%#New" -command new]
$menu add [gnocl::menuItem label -text "%#Print" \
      -command {puts [$::txtID getChars 0 end]}]
$menu add [gnocl::menuItem separator]
$menu add [gnocl::menuItem label -text "%#Quit" -command exit]

# create menu "File"
set file [gnocl::menuItem label -text "%__File" -submenu $menu]

# create menu "Edit"
set menu [gnocl::menu -children [list \
      [gnocl::menuItem label -text "%#Cut" -command cut] \
      [gnocl::menuItem label -text "%#Copy" -command copy] \
      [gnocl::menuItem label -text "%#Paste" -command paste]]]

set edit [gnocl::menuItem label -text "%__Edit" -submenu $menu]

# create menu "Help" with submenu "About" 
set help [gnocl::menuItem label -text "%__Help" -submenu \
   [gnocl::menu -children \
   [gnocl::menuItem label -text "%__About" -command about]]]

# create toolbar with icons "Quit" and "New"
set toolbar [gnocl::toolbar -style both]
$toolbar add item -text "%#Quit" -tooltip "Tooltip Quit" \
            -command exit
$toolbar add space
$toolbar add item -text "%#Cut" -tooltip "Tooltip new" \
      -command cut
$toolbar add item -text "%#Copy" -tooltip "Tooltip new" \
      -command copy
$toolbar add item -text "%#Paste" -tooltip "Tooltip new" \
      -command paste

# create text widget
set txtID [gnocl::text -editable 1]
set box [gnocl::box -orientation vertical]
$box add [gnocl::menubar -children [list $file $edit $help]] -expand 0
$box add $toolbar -expand 0
$box add $txtID

gnocl::window -title "Test of Text Widget" -child $box
## # create gnome application
## gnocl::app testApp -title "Test of Text Widget" \
##       -menubar [gnocl::menubar -children [list $file $edit $help]] \
##       -statusbar [gnocl::appBar] \
##       -toolbar $toolbar \
##       -contents $txtID

proc copy {}   { $::txtID copy }
proc cut {}    { $::txtID cut }
proc paste {}  { $::txtID paste }
proc new {}    { $::txtID remove 0 end }

proc about {} {
   puts "Text Test Application (c) 2001 - 2002 P.G. Baum"
#   gnocl::about -title "Text Test Application" -authors {{P.G. Baum}} \
#         -version [gnocl::info version] -copyright "(c) 2001 P.G. Baum"
}

set txt "The quick brown fox\njumps over\nthe lazy dog." 
$txtID insert 0 $txt
      # -foreground blue

set res [$txtID get 0 end]
if { [string compare $txt $res] } {
   puts "get all failed: \"$txt\" != \"$res\""
}
set res [$txtID get {1 6} {end-1 end}]
if { [string compare "over" $res] } {
   puts "get word failed: \"over\" != \"$res\""
}
$txtID erase {2 4} {end end-4}
set txt "The quick brown fox\njumps over\nthe dog." 
set res [$txtID get 0 end]
if { [string compare $txt $res] } {
   puts "erase failed: \"$txt\" != \"$res\""
}

$txtID insert {end end-4} "big "
set txt "The quick brown fox\njumps over\nthe big dog." 
set res [$txtID get 0 end]
if { [string compare $txt $res] } {
   puts "insert failed: \"$txt\" != \"$res\""
}

$txtID tag create monospace -family monospace
$txtID insert end "\n\nCharacter Table"
set txt ""
set noRows 64
for { set k 0 } { $k < 15 } { incr k } { 
   append txt [format "\n%3x: " [expr {$k*$noRows}]]
   for { set m 0 } { $m < $noRows } { incr m } {
      set val [expr {$k*$noRows+$m}]
      if { $val < 32 } {
         append txt " "
      } else {
         eval append txt [format {\u%x} $val] 
      }
   }
}
$txtID insert end $txt -tags monospace
$txtID insert end "\n"

# justification needs wrapMode != none
$txtID tag create justificationRight -justification right -wrapMode char
$txtID tag create justificationCenter -justification center -wrapMode word
$txtID insert end "\nJustification right\n" -tags justificationRight
$txtID insert end "Justification center\n" -tags justificationCenter

$txtID tag create superscript -size 8 -rise 10
$txtID tag create subscript -size 8 -rise -10
$txtID insert end "This is"
$txtID insert end "superscript" -tags superscript
$txtID insert end "and this"
$txtID insert end "subscript" -tags subscript
$txtID insert end "\n"
foreach val {xx-small x-small small medium large x-large xx-large } {
   set tag "tag-scale-$val"
   $txtID tag create $tag -scale $val -wrapMode word
   $txtID insert end "scale $val " -tags $tag
}
$txtID insert end "\n"

foreach {el val} {foreground blue background red font "Courier 15"
      editable no underline single underline double strikethrough on } {
   # remove spaces from tag
   set tag [string map {{ } {}} "tag-$el-$val"]
   $txtID tag create $tag -$el $val
   $txtID insert end "$el $val\n" -tags $tag
}

foreach {el val} {wrapMode word wrapMode char wrapMode none } {
   # remove spaces from tag
   set tag [string map {{ } {}} "tag-$el-$val"]
   $txtID tag create $tag -$el $val
   $txtID insert end "\n$el $val: this is a test for the wrap mode.\
         What happens to lines, which do not fit on the screen anymore?\n" \
         -tags $tag
}

puts "Automatic tests done."


if { [info exists tk_patchLevel] } {
   wm withdraw .
} else {
   gnocl::mainLoop
}
