 
/*
 * gnoclText.c --
 *
 * This file implements the text widget
 *
 * Copyright (c) 2001 P.G. Baum
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
   History:
   2002-05: transition to gtk 2.0: move from gtkText to gtkTextView
   2001-06: Begin of developement
 */

#include "gnocl.h"
#include <ctype.h>
#include <string.h>
#include <assert.h>


typedef struct
{
   GNOCL_STD_OPTIONS_VARS;

   Gnocl_BoolBool    editable;
   Gnocl_BoolBool    wordWrap;
   Gnocl_BoolBool    lineWrap;
   Gnocl_ObjBool     scrollbar;
   /* Gnocl_ObjBool  defaultFont */
} TextOptions;

static const GnoclWidgetOptions widgetOptions[] = 
{
   GNOCL_STD_OPTIONS_NAMES( TextOptions ),

   { "-editable", GNOCL_BOOL, GNOCL_OFFSET( TextOptions, editable ) },
   { "-wordWrap", GNOCL_BOOL, GNOCL_OFFSET( TextOptions, wordWrap ) },
   { "-lineWrap", GNOCL_BOOL, GNOCL_OFFSET( TextOptions, lineWrap ) },
   { "-scrollbar", GNOCL_OBJ, GNOCL_OFFSET( TextOptions, scrollbar ) },
   { NULL, 0, 0 }
};

typedef struct
{
   char              *name;
   GtkTextView       *text;
   GtkScrolledWindow *scrollWin;
   Tcl_Interp        *interp;
} TextParams;

static void freeOpts( TextOptions *p )
{
   GNOCL_STD_OPTIONS_FREE( p );

   GNOCL_FREE_TCL_OPT( p->scrollbar );
}


#if 0
static char *textTraceFunc( ClientData data,
      Tcl_Interp *interp, char *name1, char *name2, int flags)
{
   TextParams *p = (TextParams *)data;
   /* 
   Tcl_Obj *obj = Tcl_GetVar2Ex( interp, name1, name2, 0 ); 
   if( obj != NULL )
      gtk_text_set_text( GTK_ENTRY( p->text ), Tcl_GetString( obj ) );
   */

   char *txt = name1 ? Tcl_GetVar2( interp, name1, name2, 0 ): NULL;
   assert( p->interp == interp );

   if( txt != NULL )
      gtk_text_set_text( GTK_ENTRY( p->text ), txt );

   return NULL;
}

static void textChangedFunc ( GtkWidget *text, gpointer data )
{
   TextParams *p = (TextParams *)data;
   char *val = gtk_text_get_text( GTK_ENTRY( p->text ) ); 
   Tcl_SetVar( p->interp, p->variable, val, 0 );
}
#endif

static void textDestroyFunc ( GtkWidget *text, gpointer data )
{
   TextParams *p = (TextParams *)data;

   gnoclForgetWidgetFromName( p->name );
   Tcl_DeleteCommand( p->interp, p->name );

/*
   if( p->variable != NULL )
      Tcl_UntraceVar( p->interp, p->variable, TCL_TRACE_WRITES, 
            textTraceFunc, p );

   g_free( p->variable );
*/
   g_free( p->name );
   g_free( p );
}

static int textConfigure( Tcl_Interp *interp, TextParams *para,
      TextOptions *popt )
{
   GNOCL_STD_OPTIONS_HANDLE( popt, GTK_WIDGET( para->text ) );

#if 0
   if( popt->variable.changed )
   {
      if( para->variable )
      {
         Tcl_UntraceVar( para->interp, para->variable, TCL_TRACE_WRITES, 
               textTraceFunc, para );
         g_free( para->variable );
      }
      para->variable = popt->variable.val;
      popt->variable.val = NULL;

      /*
      if( para->variable == NULL )
         para->variable = g_memdup( para->name, strlen( para->name ) + 1 );
      */

      if( para->variable != NULL )
      {
         if( Tcl_TraceVar( interp, para->variable, TCL_TRACE_WRITES, 
               textTraceFunc, para ) != TCL_OK )
            return TCL_ERROR;
      }
   }

   if( popt->maxLength.changed )
      gtk_text_set_max_length( para->text, popt->maxLength.val );
#endif   

   if( popt->editable.changed )
      gtk_text_view_set_editable( para->text, popt->editable.val );

   #if 0
   if( popt->wordWrap.changed )
      gtk_text_set_editable( para->text, popt->wordWrap.val );

   if( popt->lineWrap.changed )
      gtk_text_set_editable( para->text, popt->lineWrap.val );
   #endif
   
   if( popt->scrollbar.changed )
   {
      GtkPolicyType hor, vert;
      if( gnoclGetScrollbarPolicy( interp, popt->scrollbar.val, 
            &hor, &vert ) != TCL_OK )
         return TCL_ERROR;

      gtk_scrolled_window_set_policy( para->scrollWin, hor, vert );

   #if 0
      if( popt->scrollbar.val )
      {
         if( para->box == NULL )
         {
            GtkWidget *vscrollbar = gtk_vscrollbar_new( para->text->vadj );
            /* use gtk_text_set_adjustments () ? */
            para->box =  GTK_BOX( gtk_hbox_new( 0, 0 ) );
            gtk_box_pack_start( para->box, GTK_WIDGET( para->text ), 
                  1, 1, 0 );
            gtk_box_pack_start( para->box, vscrollbar, 0, 0, 0 );
         }
      }
      /* FIXME: what to do here? ->repack */
   #endif
   }

   /* textTraceFunc( para, interp, para->variable, NULL, 0 ); */

   return TCL_OK;
}

static int posOffset( Tcl_Interp *interp, const char *txt, int *offset )
{
   *offset = 0;
   /* pos[+-]offset */
   if( *txt == '+' || *txt == '-' )
   {
      if( sscanf( txt + 1, "%d", offset ) != 1 )
      {
         Tcl_AppendResult( interp, "invalid offset \"", txt + 1, 
               "\"", NULL );
         return TCL_ERROR;
      }
      if( *txt == '-' )
         *offset *= -1;
      for( ++txt; isdigit( *txt ); ++txt )
         ;
   }
   if( *txt )
   {
      Tcl_AppendResult( interp, "invalid appendix \"", txt, 
            "\"", NULL );
      return TCL_ERROR;
   }
   return TCL_OK;
}

static int posToIter( Tcl_Interp *interp, Tcl_Obj *obj, 
      GtkTextBuffer *buffer, GtkTextIter *iter )
{
   char errMsg[] = "Position must be eiter a list of row and column "
            "or a keyword plus offset";
   char errEndOffset[] = "offset to \"end\" must be negativ";

   int len;
   if( Tcl_ListObjLength( interp, obj, &len ) != TCL_OK || len < 1 || len > 2 )
   {
      Tcl_SetResult( interp, errMsg, TCL_STATIC );
      return TCL_ERROR;
   }

   if( len == 2 )
   {
      int idx[2];
      int isEnd[2] = { 0, 0 };
      int k;
      for( k = 0; k < 2; ++k )
      {
         Tcl_Obj *tp;
         if( Tcl_ListObjIndex( interp, obj, k, &tp ) != TCL_OK )
         {
            Tcl_SetResult( interp, errMsg, TCL_STATIC );
            return TCL_ERROR;
         }
         if( Tcl_GetIntFromObj( NULL, tp, idx + k ) != TCL_OK )
         {
            char *txt = Tcl_GetString( tp );
            if( strncmp( txt, "end", 3 ) == 0 )
            {
               if( posOffset( interp, txt + 3, idx + k ) != TCL_OK )
                  return TCL_ERROR;
               if( idx[k] > 0 )
               {
                  Tcl_SetResult( interp, errEndOffset, TCL_STATIC );
                  return TCL_ERROR;
               }
               isEnd[k] = 1;
            }
            else
            {
               Tcl_AppendResult( interp, "unknown row or column index \"", 
                     txt, "\" must be integer or end plus offset" );
               return TCL_ERROR;
            }
               
         }
      }

      gtk_text_buffer_get_start_iter( buffer, iter );
      if( isEnd[0] )
      {
         gtk_text_iter_set_line( iter, -1 );
         gtk_text_iter_backward_lines( iter, -idx[0] );
      }
      else
         gtk_text_iter_set_line( iter, idx[0] );

      if( isEnd[0] )
      {
         gtk_text_iter_forward_to_line_end( iter );
         gtk_text_iter_backward_chars( iter, -idx[1] );
      }
      else
         gtk_text_iter_forward_chars( iter, idx[1] );
   }
   else if( Tcl_GetIntFromObj( NULL, obj, &len ) == TCL_OK )
   {
      if( len < 0 )
      {
         Tcl_SetResult( interp, "character offset must be greater zero.", 
               TCL_STATIC );
         return TCL_ERROR;
      }
      gtk_text_buffer_get_iter_at_offset( buffer, iter, len );
   }
   else
   {
      const char *txt = Tcl_GetString( obj );
      const char *last;
      int offset;

      if( strncmp( txt, "end", 3 ) == 0 )
      {
         gtk_text_buffer_get_end_iter( buffer, iter );
         last = txt + 3;
      }
      else if( strncmp( txt, "cursor", 6 ) == 0 )
      {
         last = txt + 6;
         gtk_text_buffer_get_iter_at_mark( buffer, iter, 
                     gtk_text_buffer_get_insert( buffer ) );
      }
      else if( strncmp( txt, "selectionStart", 14 ) == 0 )
      {
         GtkTextIter end;
         gtk_text_buffer_get_selection_bounds( buffer, iter, &end );
         last = txt + 14;
      }
      else if( strncmp( txt, "selectionEnd", 12 ) == 0 )
      {
         GtkTextIter start;
         gtk_text_buffer_get_selection_bounds( buffer, &start, iter );
         last = txt + 12;
      }
      else
      {
         Tcl_AppendResult( interp, "unknown index \"", txt, 
               "\", must be a list of row and column, "
               "an integer as character offset, "
               "or one of end, cursor, selectionStart, or selectionEnd", 
               NULL );
         return TCL_ERROR;
      }
      if( posOffset( interp, last, &offset ) != TCL_OK )
         return TCL_ERROR;

      if( offset > 0 )
         gtk_text_iter_backward_chars( iter, offset );
      else if( offset > 0 )
         gtk_text_iter_forward_chars( iter, -offset );
         
   }
   return TCL_OK;
}

static int gnoclGetScale( Tcl_Interp *interp, Tcl_Obj *obj, 
      double *val )
{
   char *txt[] = { "xx-small", "x-small", "small", "medium", "large", 
         "x-large", "xx-large", NULL };
   double types[] = { PANGO_SCALE_XX_SMALL, PANGO_SCALE_X_SMALL,
         PANGO_SCALE_SMALL, PANGO_SCALE_MEDIUM, PANGO_SCALE_LARGE,
         PANGO_SCALE_X_LARGE, PANGO_SCALE_XX_LARGE };

   if( Tcl_GetDoubleFromObj( NULL, obj, val ) != TCL_OK )
   {
      int idx;
      if( Tcl_GetIndexFromObj( NULL, obj, txt, NULL, TCL_EXACT, &idx ) 
            != TCL_OK )
      {
         Tcl_AppendResult( interp, "Unknown scale \"", Tcl_GetString( obj ),
               "\", must be a double or one of xx-small, x-small, small, "
               "medium, large, x-large, or xx-large.", NULL );
         return TCL_ERROR;
      }

      *val = types[idx];
   }

   return TCL_OK;
}

static int gnoclGetUnderline( Tcl_Interp *interp, Tcl_Obj *obj, 
      PangoUnderline *val )
{
   char *txt[] = { "none", "single", "double", "low", NULL };
   GtkUpdateType types[] = { PANGO_UNDERLINE_NONE, PANGO_UNDERLINE_SINGLE,
         PANGO_UNDERLINE_DOUBLE, PANGO_UNDERLINE_LOW };

   int idx;
   if( Tcl_GetIndexFromObj( interp, obj, txt, "underline",
         TCL_EXACT, &idx ) != TCL_OK )
      return TCL_ERROR;

   *val = types[idx];

   return TCL_OK;
}

static int gnoclGetFontDesc( Tcl_Interp *interp, Tcl_Obj *obj, 
      PangoFontDescription **desc )
{
   *desc = pango_font_description_from_string( Tcl_GetString( obj ) );

   return TCL_OK;
}

static int tagCmd( TextParams *para, Tcl_Interp *interp, int objc, 
      Tcl_Obj * const objv[] )
{
   static char *cmds[] = { "create", "configure", "add", "delete", 
         NULL };
   enum cmdIdx { CreateIdx, ConfigureIdx, AddIdx, DeleteIdx };

   typedef struct
   {
      Gnocl_ObjBool font;
      Gnocl_ObjBool foreground;
      Gnocl_ObjBool background;
      Gnocl_ObjBool wrapMode;
      Gnocl_ObjBool justification;
      Gnocl_ObjBool underline;
      Gnocl_BoolBool strikethrough;
      Gnocl_BoolBool invisible;
      Gnocl_BoolBool editable;
      Gnocl_IntBool rise;
      Gnocl_IntBool size;
      Gnocl_ObjBool scale;
      Gnocl_StringBool family;
   } TagOptions;

   static const GnoclWidgetOptions tagOptions[] = {
      { "-font", GNOCL_OBJ, GNOCL_OFFSET( TagOptions, font ) },
      { "-foreground", GNOCL_OBJ, GNOCL_OFFSET( TagOptions, 
            foreground ) },
      { "-background", GNOCL_OBJ, GNOCL_OFFSET( TagOptions, 
            background ) },
      { "-wrapMode", GNOCL_OBJ, GNOCL_OFFSET( TagOptions, wrapMode ) },
      { "-justification", GNOCL_OBJ, GNOCL_OFFSET( TagOptions, 
            justification ) },
      { "-underline", GNOCL_OBJ, GNOCL_OFFSET( TagOptions, underline ) },
      { "-strikethrough", GNOCL_BOOL, GNOCL_OFFSET( TagOptions, 
            strikethrough ) },
      { "-invisible", GNOCL_BOOL, GNOCL_OFFSET( TagOptions, invisible ) },
      { "-editable", GNOCL_BOOL, GNOCL_OFFSET( TagOptions, editable ) },
      { "-rise", GNOCL_INT, GNOCL_OFFSET( TagOptions, rise ) },
      { "-size", GNOCL_INT, GNOCL_OFFSET( TagOptions, size ) },
      { "-scale", GNOCL_OBJ, GNOCL_OFFSET( TagOptions, scale ) },
      { "-family", GNOCL_STRING, GNOCL_OFFSET( TagOptions, family ) },
      { NULL, 0, 0 }
   };

   int idx;
   GtkTextBuffer *buffer = gtk_text_view_get_buffer( para->text );

   if( Tcl_GetIndexFromObj( interp, objv[2], cmds, "subcommand", 
         TCL_EXACT, &idx ) != TCL_OK )
      return TCL_ERROR;

   switch( idx )
   {
      case CreateIdx:
            {
               GtkTextTag *tag; 
               TagOptions opt;
               /* win tag create */
               if( objc < 4 )
               {
                  Tcl_WrongNumArgs( interp, 2, objv, 
                        "tag-name ?option val ...?" );
                  return TCL_ERROR;
               }
               /* tag = gtk_text_tag_new( Tcl_GetString( objv[3] ) ); */
               tag = gtk_text_buffer_create_tag( buffer, 
                     Tcl_GetString( objv[3] ), NULL );

               memset( &opt, 0, sizeof( opt ) );
               if( gnoclParseAllWidgetOpts( interp, objc, objv, 3,
                     tagOptions, (char *)&opt ) != TCL_OK )
                  return TCL_ERROR;
               if( opt.font.changed )
               {
                  PangoFontDescription *font;
                  int error = 1;
                  if( gnoclGetFontDesc( interp, opt.font.val, &font) 
                        == TCL_OK )
                  {
                     g_object_set( (gpointer)tag, "font-desc", font, NULL );
                     error = 0;
                  }
                  pango_font_description_free( font );
                  if( error )
                     return TCL_ERROR;
               }
               if( opt.foreground.changed )
               {
                  GdkColor color;
                  if( gnoclGetGdkColor( interp, opt.foreground.val, &color,
                        NULL ) != TCL_OK )
                     return TCL_ERROR;
                  g_object_set( (gpointer)tag, "foreground-gdk", &color, NULL );
               }
               if( opt.background.changed )
               {
                  GdkColor color;
                  if( gnoclGetGdkColor( interp, opt.background.val, &color,
                        NULL ) != TCL_OK )
                     return TCL_ERROR;
                  g_object_set( (gpointer)tag, "background-gdk", &color, NULL );
               }
               if( opt.wrapMode.changed )
               {
                  GtkWrapMode val;
                  if( gnoclGetWrapMode( interp, opt.wrapMode.val, &val ) 
                        != TCL_OK )
                     return TCL_ERROR;
                  g_object_set( (gpointer)tag, "wrap-mode", val, NULL );
               }
               if( opt.justification.changed )
               {
                  GtkJustification val;
                  if( gnoclGetJustification( interp, opt.justification.val, 
                        &val ) != TCL_OK )
                     return TCL_ERROR;
                  g_object_set( (gpointer)tag, "justification", val, NULL );
               }
               if( opt.underline.changed )
               {
                  PangoUnderline ul;
                  if( gnoclGetUnderline( interp, opt.underline.val, &ul ) 
                        != TCL_OK )
                     return TCL_ERROR;
                  g_object_set( (gpointer)tag, "underline", ul, NULL );
               }
               if( opt.strikethrough.changed )
                  g_object_set( (gpointer)tag, "strikethrough", 
                        opt.strikethrough.val, NULL );
               if( opt.invisible.changed )
                  g_object_set( (gpointer)tag, "invisible", 
                        opt.invisible.val, NULL );
               if( opt.editable.changed )
                  g_object_set( (gpointer)tag, "editable", opt.editable.val, 
                        NULL );
               if( opt.rise.changed )
                  g_object_set( (gpointer)tag, "rise", 
                        opt.rise.val * PANGO_SCALE, NULL );
               if( opt.size.changed )
                  g_object_set( (gpointer)tag, "size", 
                        opt.size.val * PANGO_SCALE, NULL );
               if( opt.scale.changed )
               {
                  double val;
                  if( gnoclGetScale( interp, opt.scale.val, &val ) 
                        != TCL_OK )
                     return TCL_ERROR;
                  g_object_set( (gpointer)tag, "scale", val, NULL );
               }
               if( opt.family.changed )
                  g_object_set( (gpointer)tag, "family", 
                     opt.family.val, NULL );
            }
            return TCL_OK;
      default:
            assert( 0 );
            return TCL_ERROR;
   }

   return TCL_OK; 
}

static int textInsert( TextParams *para, Tcl_Interp *interp, int objc, 
      Tcl_Obj * const objv[] )
{
   typedef struct
   {
      Gnocl_ListBool tags;
   } TextInsertOptions;
   static const GnoclWidgetOptions insertOptions[] = {
      { "-tags", GNOCL_LIST, GNOCL_OFFSET( TextInsertOptions, tags ) },
      { NULL, 0, 0 }
   };
   TextInsertOptions opt;
   gint              startOffset;

   GtkTextBuffer *buffer = gtk_text_view_get_buffer( para->text );
   GtkTextIter   iter;

   if( objc < 4 )
   {
      Tcl_WrongNumArgs( interp, 2, objv, "position text ?-option val ...?" );
      return TCL_ERROR;
   }

   if( posToIter( interp, objv[2], buffer, &iter ) != TCL_OK )
      return TCL_ERROR;

   memset( &opt, 0, sizeof( opt ) );
   if( gnoclParseAllWidgetOpts( interp, objc, objv, 3,
         insertOptions, (char *)&opt ) != TCL_OK )
      return TCL_ERROR;

   startOffset = gtk_text_iter_get_offset( &iter );

   gtk_text_buffer_insert( buffer, &iter, gnoclGetString( objv[3] ), -1 );
   if( opt.tags.changed )
   {
      GtkTextIter start;
      int k, no;

      gtk_text_buffer_get_iter_at_offset( buffer, &start, startOffset );
      if( Tcl_ListObjLength( interp, opt.tags.val, &no ) != TCL_OK )
         return TCL_ERROR;

      for( k = 0; k < no; ++k )
      {
         Tcl_Obj *tp;
         if( Tcl_ListObjIndex( interp, opt.tags.val, k, &tp ) != TCL_OK )
         {
            Tcl_SetResult( interp, "Could not read tag list", TCL_STATIC );
            return TCL_ERROR;
         }
         gtk_text_buffer_apply_tag_by_name( buffer, Tcl_GetString( tp ), 
               &start, &iter );
      }
   }

#if 0
   /* widget insert pos text ?-option val ...? */
   typedef struct
   {
      Gnocl_ObjBool font;
      Gnocl_ObjBool foreground;
      Gnocl_ObjBool background;
   } TextInsertOptions;

   static const GnoclWidgetOptions insertOptions[] = {
      { "-font", GNOCL_OBJ, GNOCL_OFFSET( TextInsertOptions, font ) },
      { "-foreground", GNOCL_OBJ, GNOCL_OFFSET( TextInsertOptions, 
            foreground ) },
      { "-background", GNOCL_OBJ, GNOCL_OFFSET( TextInsertOptions, 
            background ) },
      { NULL, 0, 0 }
   };

   TextInsertOptions opt;
   GdkFont *font = NULL;
   GdkColor foreground, background;
   int      index, oldIndex;


   if( objc < 4 )
   {
      Tcl_WrongNumArgs( interp, 2, objv, "position text ?option val ...?" );
      return TCL_ERROR;
   }

   if( gnoclEditablePosToIndex( interp, objv[2], GTK_EDITABLE( para->text ), 
         &index ) != TCL_OK )
      return TCL_ERROR;

   memset( &opt, 0, sizeof( opt ) );
   if( gnoclParseAllWidgetOpts( interp, objc, objv, 3,
         insertOptions, (char *)&opt ) != TCL_OK )
      return TCL_ERROR;
   
   if( opt.font.changed )
   {
      if( gnoclGetGdkFont( interp, opt.font.val, &font ) != TCL_OK )
         return TCL_ERROR;
   }
   else     /* TODO: how to get the default font? */
      font = gdk_font_load( 
            "-adobe-courier-bold-r-normal-*-*-140-*-*-m-*-iso8859-1" );

   if( opt.foreground.changed )
   {
      if( gnoclGetGdkColor( interp, opt.foreground.val, &foreground, NULL ) )
         return TCL_ERROR;
   }
   else
      gdk_color_black( gtk_widget_get_colormap( GTK_WIDGET( para->text ) ), 
            &foreground );

   if( opt.background.changed )
   {
      if( gnoclGetGdkColor( interp, opt.background.val, &background, NULL ) )
         return TCL_ERROR;
   }
   else
      gdk_color_white( gtk_widget_get_colormap( GTK_WIDGET( para->text ) ), 
            &background );

   oldIndex = gtk_text_get_point( para->text );
   gtk_text_set_point( para->text, index );

   gtk_text_insert( para->text, font, &foreground, &background,
      gnoclGetString( objv[3] ), -1 );

   gtk_text_set_point( para->text, oldIndex );
#endif

   return TCL_OK;
}

static int textFunc( ClientData data, Tcl_Interp *interp,
      int objc, Tcl_Obj * const objv[] )
{
   /* TODO: posForward, posWordBackward, posIsEnd ... */
   static char *cmds[] = { GNOCL_STD_CMD_NAMES,
         /* "erase" means "delete", but this is already used by
            the widget std commands */
         "erase", "select", 
         "cut", "copy", "paste", "get", "getLength",
         "configure", 
         "setCursor", "getCursor", "getPosition", "insert",
         "tag",
         NULL };
   enum cmdIdx { GNOCL_STD_CMD_ENUMS, 
         EraseIdx, SelectIdx, 
         CutIdx, CopyIdx, PasteIdx, GetIdx, GetLengthIdx,
         ConfigureIdx,
         SetCursorIdx, GetCursorIdx, GetPositionIdx, InsertIdx,
         TagIdx };
   TextParams *para = (TextParams *)data;
   GtkWidget *text = GTK_WIDGET( para->text );
   GtkTextBuffer *buffer = gtk_text_view_get_buffer( para->text );
   int idx;

   if( Tcl_GetIndexFromObj( interp, objv[1], cmds, "command", 
         TCL_EXACT, &idx ) != TCL_OK )
      return TCL_ERROR;

   switch( idx )
   {
      GNOCL_STD_CMD_CASE( interp, objc, objv, text );
            break;
      case EraseIdx:
      case SelectIdx:
      case GetIdx:
            {
               GtkTextIter startIter, endIter;
               /* text erase/select/getChars startIndex ?endIndex? */
               if( objc < 4 || objc > 4 )
               {
                  Tcl_WrongNumArgs( interp, 2, objv, "startIndex ?endIndex?" );
                  return TCL_ERROR; 
               }
               if( posToIter( interp, objv[2], buffer, &startIter ) != TCL_OK )
                  return TCL_ERROR;

               if( objc >= 4 )
               {
                  if( posToIter( interp, objv[3], buffer, 
                        &endIter ) != TCL_OK )
                     return TCL_ERROR;
               }
               else
               {
                  endIter = startIter;
                  gtk_text_iter_backward_char( &endIter );
               }
               
               switch( idx )
               {
                  case EraseIdx:
                           gtk_text_buffer_delete( buffer, 
                                 &startIter, &endIter );
                           break;
                  case SelectIdx:
                           gtk_text_buffer_place_cursor( buffer, &startIter );
                           gtk_text_buffer_move_mark_by_name( buffer, 
                                 "selection_bound", &endIter );
                           break;
                  case GetIdx:
                           {
                              /* TODO: include_hidden_chars */
                              char *txt = gtk_text_buffer_get_text(
                                    buffer, &startIter, &endIter, 1 );
                              Tcl_SetObjResult( interp, 
                                    Tcl_NewStringObj( txt, -1 ) );
                           }
                           break;
               }
            }
            break;
      case CutIdx:
      case CopyIdx:
      case PasteIdx:
            {
               /* TODO: option which clipboard */
               GtkClipboard *clipboard = gtk_clipboard_get( GDK_NONE );
               if( objc != 2 )
               {
                  Tcl_WrongNumArgs( interp, 2, objv, "" );
                  return TCL_ERROR; 
               }
               switch( idx )
               {
                  case CutIdx:
                     gtk_text_buffer_cut_clipboard( buffer, clipboard, 1 );
                     break;
                  case CopyIdx:
                     gtk_text_buffer_copy_clipboard( buffer, clipboard ); 
                     break;
                  case PasteIdx:
                     gtk_text_buffer_paste_clipboard( buffer, clipboard,
                           NULL, 1 ); 
                     break;
               }
            }
            break;
      case GetLengthIdx: /* TODO getByteCount */
            {
               /* editable getLength */
               if( objc != 2 )
               {
                  Tcl_WrongNumArgs( interp, 2, objv, "" );
                  return TCL_ERROR; 
               }

               Tcl_SetObjResult( interp, Tcl_NewIntObj( 
                     gtk_text_buffer_get_char_count( buffer ) ) );
            }
            break;

      case ConfigureIdx:
            {
               TextOptions opt;
               int ret = TCL_ERROR;

               memset( &opt, 0, sizeof( opt ) );
               if( gnoclParseAllWidgetOpts( interp, objc, objv, 1,
                     widgetOptions, (char *)&opt ) == TCL_OK )
                  ret = textConfigure( interp, para, &opt );

               freeOpts( &opt );
               if( ret != TCL_OK )
                  return TCL_ERROR;
            }
            break;
      case SetCursorIdx:
            {
               GtkTextIter iter;

               /* text setCursor index */
               if( objc != 3 )
               {
                  Tcl_WrongNumArgs( interp, 2, objv, "index" );
                  return TCL_ERROR; 
               }
               if( posToIter( interp, objv[2], buffer, &iter ) != TCL_OK )
                  return TCL_ERROR;

               gtk_text_buffer_place_cursor( buffer, &iter );
            }
            break;
      case GetCursorIdx:
            {
               GtkTextIter   iter;
               int           row, col;
               Tcl_Obj       *resList;

               /* text getCursor */
               if( objc != 2 )
               {
                  Tcl_WrongNumArgs( interp, 2, objv, "" );
                  return TCL_ERROR; 
               }
               gtk_text_buffer_get_iter_at_mark( buffer, &iter,
                     gtk_text_buffer_get_insert( buffer ) );

               row = gtk_text_iter_get_line( &iter );
               col = gtk_text_iter_get_line_offset( &iter );

               resList = Tcl_NewListObj( 0, NULL );
               Tcl_ListObjAppendElement( interp, resList,
                     Tcl_NewIntObj( row ) );
               Tcl_ListObjAppendElement( interp, resList,
                     Tcl_NewIntObj( col ) );
               Tcl_SetObjResult( interp, resList );
            }
            break;
#if 0
      case GetPositionIdx:
            {
               int index;
               /* text getPosition pos */
               if( objc != 3 )
               {
                  Tcl_WrongNumArgs( interp, 2, objv, "position" );
                  return TCL_ERROR; 
               }
               if( gnoclEditablePosToIndex( interp, objv[2], 
                     GTK_EDITABLE( para->text ), &index ) != TCL_OK )
                  return TCL_ERROR;
               Tcl_SetObjResult( interp, Tcl_NewIntObj( index ) );
            }
            break;
#endif
      case InsertIdx:
            return textInsert( para, interp, objc, objv );
      case TagIdx:
            return tagCmd( para, interp, objc, objv );
      default:
            assert( 0 );
            return TCL_ERROR;
   }

   return TCL_OK;
}

int gnoclTextCmd( ClientData data, Tcl_Interp *interp,
      int objc, Tcl_Obj * const objv[] )
{
   TextOptions opt;
   TextParams  *para;
   GtkWidget   *widget;

   memset( &opt, 0, sizeof( opt ) );
   if( gnoclParseAllWidgetOpts( interp, objc, objv, 0,
         widgetOptions, (char *)&opt ) != TCL_OK )
      return TCL_ERROR;

   para = g_new( TextParams, 1 );
   para->text = GTK_TEXT_VIEW( gtk_text_view_new( ) );
   para->name = gnoclGetAutoWidgetId();
   para->interp = interp;
   /* para->variable = NULL; */

   para->scrollWin = GTK_SCROLLED_WINDOW( 
         gtk_scrolled_window_new( NULL, NULL ) );
   gtk_scrolled_window_set_policy( para->scrollWin, 
         GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
   gtk_container_add( GTK_CONTAINER( para->scrollWin ), 
         GTK_WIDGET( para->text ) );

   if( textConfigure( interp, para, &opt ) != TCL_OK )
   {
      freeOpts( &opt );
      /* FIXME: g_free para->name ... */
      gtk_widget_destroy( GTK_WIDGET( para->scrollWin ) );
      gtk_widget_destroy( GTK_WIDGET( para->text ) );
      return TCL_ERROR;
   }
   freeOpts( &opt );

   widget = GTK_WIDGET( para->scrollWin ); 
/*
   gtk_signal_connect( GTK_OBJECT( para->text ), "changed", 
         GTK_SIGNAL_FUNC( textChangedFunc ), para );
*/
   gtk_signal_connect( GTK_OBJECT( widget ), "destroy", 
         GTK_SIGNAL_FUNC( textDestroyFunc ), para );

   gnoclMemNameAndWidget( para->name, widget );

   gtk_widget_show_all( GTK_WIDGET( widget) );

   Tcl_CreateObjCommand( interp, para->name, textFunc, para, NULL );

   Tcl_SetObjResult( interp, Tcl_NewStringObj( para->name, -1 ) );

   return TCL_OK;
}

